# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, List, Type

from .base import BasePipeline
from .evaluation import EvaluationPipeline
from .inference import InferencePipeline
from .task_processors.base import BaseTaskProcessor
from .task_processors.classification import ClassificationProcessor
from .task_processors.detection import DetectionProcessor
from .task_processors.pose_estimation import PoseEstimationProcessor
from .task_processors.segmentation import SegmentationProcessor
from .train import TrainingPipeline

# TODO: Temporary defined. It should be integrated with `..models.registry.SUPPORTING_TASK_LIST`
SUPPORTING_TASK_LIST: List[str] = ['classification', 'segmentation', 'detection', 'pose_estimation']

TASK_PROCESSOR: Dict[str, Type[BaseTaskProcessor]] = {
    'classification': ClassificationProcessor,
    'segmentation': SegmentationProcessor,
    'detection': DetectionProcessor,
    'pose_estimation': PoseEstimationProcessor,
}

PIPELINES: Dict[str, Type[BasePipeline]] = {
    'train': TrainingPipeline,
    'evaluation': EvaluationPipeline,
    'inference': InferencePipeline,
}
